@rem = '
@goto endofperl
';

########################################################################
#
# Sept 18, 1996  Created	ReenaA
# Jan  06, 1997  Modified	CostelR
#
# Usage :
#	CInf file1.inf [file2.inf] [file3.inf] ...
# Output:
#	.\htm\file1.htm
#	....
#	.\htm\summary.htm
#
# SYNOPSIS: Check INF files. It generates an html report to stdout.
########################################################################
		# INF procedures indexed by ClassGUID
		# the entry value is a string of names separated by commas :
		# '<Perl_Module_to_Load>, <INF_Class_Name>, <Checking_Procedure>,
		# <Init_Globals_Procedure>, <Summary_Procedure>'
		# all of these are optional, except for <Pearl_Module_to_Load>

	$CInfINI = "CINF.INI";
	%INFCheckProc =();
 
########################################################################
#
# Errors string table
#
########################################################################
%ErrorTable = (
	1001 => ["Incorrect section definition."],
	1002 => ["Section redefined. [%s] already defined at line %d. Not processing this section."],
	1003 => ["Section [%s] not defined."],
	1004 => ["Entry value should be in quotes."],
	1005 => ["String has more than %d characters."],
	1006 => ["Don't know to process %s keyword or redefined key."],
	1007 => ["Provider %s not found in [STRINGS] section."],
	1008 => ["Signature is expected to be \"\$CHICAGO\$\""],
	1009 => ["Class is expected to be \"%s\"."],
	1010 => ["ClassGUID is expected to be \"%s\"."],
	1011 => ["Key %s not defined."],
	1012 => ["Manufacturer %s not found in [STRINGS] section."],
	1013 => ["Not more than one word on the right of \"=\"."],
	1014 => ["Illegal entry name."],
	1015 => ["Model %s not found in [STRINGS] section."],
	1016 => ["Missing Install section identifier. Not processing this model."],
	1017 => ["Too many rank ids (three commas maximum)."],
	1018 => ["Missing Rank0 ID."],
	1019 => ["%s : install sections are inconsistent."],
	1020 => ["Entry name %s illegal."],
	1021 => ["Duplicated key %s."],
	1022 => ["%s should not have a value."],
	1023 => ["Only one of the keys REBOOT and RESTART is allowed."],
	1024 => ["Only one section allowed for %s."],
	1025 => ["No value for %s."],
	1026 => ["Empty section name for %s."],
	1027 => ["Blank identifier."],
	1028 => ["Line has and odd number of \"s. Add one \" at the end of line to continue processing."],
	1029 => ["String %s not referenced."],
	1030 => ["Line is expected to be of the form \"Entry_Name = Entry_Value\"."],
	1031 => ["Not more than one word on the left of \"=\"."],
	1032 => ["Entry value is empty."],
	1033 => ["Semicolon inside comment."],
	1034 => ["File too big, maximum 62K allowed."],
	1035 => ["Identifier %s not defined."],
	1036 => ["String %s not defined in [STRINGS] section."],
	1037 => ["Section [%s] expected but not found."],
	1038 => ["Disk number %s not found in disk section [%s]."],
	1039 => ["No manufacturer defined in section [%s]."],
	1040 => ["No model defined in section [%s]."],
	1041 => ["Last value is empty, remove trailing commas."],
	1042 => ["File_section keyword redefined."],
	1043 => ["LDID value is missing."],
	1044 => ["Wrong value for LDID."],
	1045 => ["Keyword %s not defined in [DESTINATIONDIRS]."],
	1046 => ["[DESTINATIONDIRS] section is not defined."],
	1047 => ["Section not referenced by any copyfiles, renfiles or delfiles keyword."],
	1048 => ["String entry redefined."],
	1049 => ["Identifier has more than %d characters."],
	1050 => ["The inf type cannot be identified (either CINF.INI is missing or GUID entry not found)."],
	1051 => ["The quotes surrounding the IDs should be removed."],
	1052 => ["Install section %s not found."],
	1053 => ["\/ not allowed in Quoted strings"], #Added for netcard type
);

%WarningTable = (
	2001 => ["Line has more than %d characters."],
	2002 => ["Provider is expected to be of the form %%name%%."],
	2003 => ["LayoutFile is optional. If present should be %s."],
	2004 => ["Entry keyword should have the form %%Device_description_name%%"],
	2005 => ["Illegal character in string key."],
	2006 => ["Same model \"%s\", already defined under a different manufacturer."],
	2007 => ["%s doesn't require = ."],
	2008 => ["Line has an odd number of \"s. Correct this to avoid erroneous processing."],
	2009 => ["TAB inside quoted string."],
	2010 => ["Section [%s] not referenced."],
	2011 => ["White spaces in identifier \"%s\"."],
	2012 => ["Identifier %s found in a previous %s line."],
	2013 => ["LayoutFile should ONLY be used to specify cabinet files, vendors shouldn't use."],
	2014 => ["LayoutFile absent, should be used only by Microsoft engineering, not by vendors."],
	2015 => ["Entries are permitted only for floppy (vendor) installation."],
	2016 => ["Keyword LayoutFile in [VERSION] section not found."],
);

########################################################################
#
# Package with version constants for current file
#
########################################################################
{ package CINFVersion;

  $BuildVersion = "1.00.0019";
}

########################################################################

print "CINF: Copyright 1996-97 Microsoft Corporation\n";
print "      Version   $CINFVersion'BuildVersion\n\n";

require('infparse.pl');

&LoadModules();			#Load modules given in $CInfINI

&VersionChecking();

if (@ARGV==0)
{
	die "Usage: $0 <input_file_list> [<browser_name>/E]\n\n".
		"       <input_file> = file_name | [directory_name]/ALL\n".
		"       <browser_name> = html browser\n";
}

#--------------------------------------------------------------------
# Globals

InitGlobalsAll();
#--------------------------------------------------------------------

local ($Overwrite, $Browser); 		# check summary file exists

undef($Browser);

$Overwrite = "A";
opendir(HTM, "htm") || goto Okay;

$Overwrite = "Y";
		# check for .\htm\summary.htm
if (-e ".\\htm\\summary.htm")
  { print "Summary file .\\htm\\summary.htm already exists. Overwrite (Yes/No/All) [Y] ? ";
	$Overwrite = <STDIN>;
	chop $Overwrite;
  }

die "Save contents of .\\htm and restart.\n" if ($Overwrite =~ /^[nN]/);

Okay:
# get all files to process
@AllFiles = ();
foreach $infile (@ARGV)
{
	local (@Files, $dir);

	if ( (uc $infile) =~ /(.*)\/ALL$/ )
	{	$dir = (($1 =~ /^$/) ? "." : $1);
		if ( !opendir(LOCALDIR,  $dir) )
		{	print "Cannot open directory $dir \n";
			next;
		}
		@Files = readdir LOCALDIR;
		foreach $file (@Files)
		{	push(@AllFiles, (($dir =~ /^\.$/) ? "" : $dir."\\") .$file) if ((uc $file) =~ /\.INF$/); }
		@AllFiles	= sort @AllFiles;
		closedir LOCALDIR;
	}
	elsif ( (uc $infile) =~ /(.*)\/E$/ )
	{
		$Browser = $1;
	}
	else
	{	push(@AllFiles, $infile); }
}

########################################################################
#
# Main function -  executed for each input file. 
#
########################################################################
foreach $infile (@AllFiles)
{
  %TimeMarks = ();
  $IndentValue = 0;

  BeginTimer("Main");

  InitGlobalsCurrent();

  #--------------------------------------------------------------------
  # check input filename.
  #--------------------------------------------------------------------
  unless (open (INFILE, "$infile"))
  {
		print "Error opening $infile: $!\n";
		next;		 
  }
  close(INFILE);

  #--------------------------------------------------------------------
  # locals - should not get messed with in subs.
  #--------------------------------------------------------------------
  local($outfile, $key, @list, @INFProcedures);

	$NumFiles++;
	@INFProcedures = ();

	#decide outfile name.
	# Example: in.inf ==> in.htm
	$outfile = $infile;
	$outfile =~ s/^([a-zA-Z]:)?//;	# Strip off preceeding drive letter.
	$outfile =~ s/^.*\\//;			# Strip off directories
	$outfile =~ s/\.\w*$//; 		# Strip off extension
	$outfile = ".\\htm\\$outfile.htm";

	# check for the file in \htm\
	if (($Overwrite !~ /^[aA]/) && (-e $outfile) )
	{ print "Output file $outfile already exists. Overwrite (Yes/No/All) [Y] ? ";

	  $Overwrite = <STDIN>;
	  chop $Overwrite;

	  if ($Overwrite =~ /^[nN]/)
		{ $NumFiles--;
		  print "$infile skipped\n";
		  next;
		}
	}
	print "Processing $infile -> $outfile ";

	&preprocess($infile);	# gets the INF lines and store them in 'INFLines' array

	&process_sections();	# reads the INF sections and creates $SecStart and $SecEnd

	@INFProcedures = &DispatchInf();
				# identifies the INF type and calls then
				#	- INF specific initialization procedure
				#	- StrProc for [STRINGS] checking
				#	- VerProc for [VERSION] checking
				#	- INF specific checking procedure

	EndTimer("Main");
}

########################################################################
# End of processing, build the summary.htm file
########################################################################
# process all R0s collected.

if ($PrintModemR0Statistics)
   { &proc_AllR0; }

# now print summary of all files processed.
&print_main_page;

if ( defined($Browser) )
{
	print ("Starting explorer, please wait!\n");
	system("Start $Browser .\\htm\\summary.htm\n");
}
die "Done.\n";

########################################################################
#
#	Fill %INFCheckProc using the $CInfFile
#
########################################################################
sub
LoadIniFile
{
	local($Line, $Key, $Value, @Procedures);

	if( !(-e $CInfINI) )	# ini file not found
	{
		print "Cannot find $CInfINI\n";
		return;
	}
	return if ( !open (INIFILE, $CInfINI) );

	while(<INIFILE>)
	{
		$Line = $_;

		chop $Line if /\n$/;	# removes \n
		$Line =~ s/[\s\t]//g;	#strip all white spaces.
		$Line =~ s/^(.*);(.*)/$1/;	#strip comments.
		next if ($Line =~ /^$/);

		if ($Line =~ /^(.*)=(.*)$/) # keyword
		{
			$Key = uc $1;
			$Value = $2;
			if ($Key =~ /^DEFAULT$/)
			{	$INFCheckProc{$Key}	= uc $Value;
				next;
			}

			if ( $Key =~ /^{.*}$/ )	
			{
				$INFCheckProc{$Key}	= $Value;
				@Procedures = split(",",$Value);
				if (defined($Procedures[1]) &&
					$Procedures[1] !~ /^$/)				# ClassName
				{
					$INFCheckProc{uc $Procedures[1]} = $Key;
				}
			}
		}
	}

	close INIFILE;
}

########################################################################
#
#	Load modules as required by %INFCheckProc
#
########################################################################
sub
LoadModules
{
local($key, $INFModule, @Modules, $size, $i, $Answer);

	unshift(@INC, '.');

	LoadIniFile();

	while (($key, $INFModule) = each %INFCheckProc)
	{	
		next if ($key !~ /^{.*}$/);

		@Modules = split(",",$INFModule);
		$size = @Modules;
		for ($i=0; $i < $size; $i++)
			{  $Modules[$i] =~ s/[\s\t]//g; }
		if (!defined($Modules[0]) || $Modules[0] =~ /^$/)
		   { print "$key: module name is empty. Continue (Yes/No) [Y] ? "; 
			 $Answer = <STDIN>;  chop $Answer;
			die "Done.\n" if ($Answer =~ /^[nN]/);
			next;
		   }
		if ( !(-e $Modules[0]) )
		{	
			print "Module $Modules[0] not found.\n";
			next;
		}
		print "Load module $Modules[0]\n";
		require($Modules[0]);
	}
}

########################################################################
#
#	Default processing of INF file and INF specific procedure
#	(as given in %INFCheckProc)
#
########################################################################
sub
DispatchInf
{
local($line, $EntryName, $EntryValue, $CLnum, $LLnum);
local(%VersionEntries, $ClassGui, $ClassName, $INFModules, $proc, @InfSpecific, $i, $size);

	@InfSpecific = ();

	if (! defined($CLnum = $SecStart{$VerSec}) ||
	! defined($LLnum = $SecEnd{$VerSec})
	)
	{ return; }

	%VersionEntries = ();

	while(++$CLnum < $LLnum)
	{
		$line = $INFlines[$CLnum];
		if (! defined($line))
		{ next; }

		$EntryName = uc &GetEntryName($line); #Change LHS to uppercase. 
		$EntryValue = &GetEntryValue($line);
		$EntryValue = $AllQuotes{$1} if($EntryValue =~ /^__QUOTED_STR_\(([0-9]*)\)__$/);

		$VersionEntries{$EntryName} = $EntryValue;
	} 

	undef $ClassGui;
	undef $ClassName;
	undef $INFModules;
	if (!defined ($ClassGui = $VersionEntries{'CLASSGUID'}))	# Wrong ClassGUID
	{			# Try ClassName
		if (defined ($ClassName	= $VersionEntries{'CLASS'}) )
		{	$ClassGui	= $INFCheckProc{uc $ClassName}; }
	}

	if (defined($ClassGui))
	{
		$INFModules = $INFCheckProc{uc $ClassGui};
	}

	if (!defined($INFModules))			# apply defaults
	{
		$ClassGui	= $INFCheckProc{'DEFAULT'};
		$INFModules = $INFCheckProc{uc $ClassGui};
	}

	@InfSpecific = split(",", $INFModules);
	$size = @InfSpecific;
	for ($i=0; $i < $size; $i++)
	{	$InfSpecific[$i] =~ s/[\s\t]//g; }

	if (defined($InfSpecific[1]))
	{	print "[$InfSpecific[1] type] "; }
	else
	{	print "[Unknown type] "; }

			# [0] = perl file to include
			# [1] = INF class name
			# [2] = Check INF specific procedure
			# [3] = Init INF specific variables
			# [4] = Summary INF specific procedure

	if (defined($InfSpecific[3]) && 	# call INIT specific variables
	$InfSpecific[3] !~ /^$/)
	{
		$proc = $InfSpecific[3];
		&$proc;
	}

	#
	# process the compulsory sections.
	#

	&StrProc($StrSec);			# [STRINGS] section

	$VerVal{'CLASS'} = (defined($InfSpecific[1])) ? $InfSpecific[1] : "";
	$VerVal{'CLASSGUID'} = (defined($ClassGui)) ? $ClassGui : "";
	if (!defined($ClassGui) || !defined($InfSpecific[1]))
	{
		add_ref_err(1050, $SecStart{$VerSec});
	}

	&VerProc($VerSec);			# call [VERSION] section checking procedure

	&ManuProc($ManuSec);		# [MANUFACTURER] section

	&CFlagProc($CFlagSec);		# [CONTROLFLAGS] section

	&DestDirs($DestDirsSec);	# [DESTINATIONDIRS] section

	foreach $manu (keys(%ManuList)) # Find models for each manufacturer
	{
		&ModelFind($manu);
	}

	# and process the install sectoins.
	foreach $InstSection (keys %InstList)
	{
		&InstProc($InstSection);
		if (defined($SecStart{$InstSection.'.HW'}))		# default section used to add hardware registry keys 
		{	&InstProc($InstSection.'.HW', 1);	}
	}

	&SourceDisks();

	if (defined($InfSpecific[2]) && # call INF specific checking procedure
	$InfSpecific[2] !~ /^$/)
	{ $proc = $InfSpecific[2];
		  &$proc;
	}

	print "\n"; 	# after ...

	&String_usage();
	&Section_usage();
	&ExcludeSelect_usage();
					# print result in .htm file
	&print_infile_and_annotations($infile, $outfile, @InfSpecific);
	$NumErrFiles++ if ($NumErrs);

	@InfSpecific;	# return value
}

#------------------------------------------------------------
# Adds the infile processing summary to Sumary
#
# Globals changed:
#		Summary
#
# Usage: 
#		summarize(infile_name, outfile_name)
#------------------------------------------------------------

sub
summarize
{
local($infile, $outfile) = @_;

	$outfile =~ s/^.*\\//;			# Strip off directories

	#add summary for this file to %Summary.
	($NumErrs) ? ($Summary{$infile} = join( "",
				"<td><FONT COLOR=RED><PRE><font size=+2>",
				"<b>$infile\t\t</b>",
				"</FONT></pre></td><td><pre>",
				"  <A HREF=\"$outfile#Errors\">",
				"Errors   : $NumErrs",
				"</A></PRE></td></FONT>\n"))
		: (($NumWrns) ? ($Summary{$infile} =  join( "",
				"<td><FONT COLOR=BLUE><PRE><font size=+2>",
				"<b>$infile\t\t</b>",
				"</FONT></pre></td><td><pre>",
				"  <A HREF=\"$outfile#Warnings\">",
				"Warnings : $NumWrns",
				"</A></PRE></td></FONT>\n"))
					  : ($Summary{$infile} =  join( "", 
						"<td><FONT COLOR=GREEN><PRE><font size=+2>",
						"<b>$infile\t\t</b>",
						"</FONT></pre></td><td><pre>",
						"  <A HREF=\"$outfile#Top\">",
						"Checked OK",
						"</A></PRE></td></FONT>\n")));
	if ($NumErrs > 0 && $NumWrns > 0)
	{ $Summary{$infile} = join( "", $Summary{$infile}, "\n", 
				"<tr><td></td><td><pre>",
				"  <A HREF=\"$outfile#Warnings\">",
				"Warnings : $NumWrns",
				"</A></PRE></td></tr>\n");
	}
}

#------------------------------------------------------------
# Prints the output file in the desired format.
#
# Globals changed:
#
# Usage: 
#		print_infile_and_annotations(infile_name, outfile_name, Inf_Specific_Procedures)
#------------------------------------------------------------

sub 
print_infile_and_annotations
{
local($infile, $outfile, @INFProcedures) = @_;
local($oldhandle, $title, $SummaryProcedure, $INFClass); 

$INFClass		= $INFProcedures[1];
$SummaryProcedure	= $INFProcedures[4];

	# Create some formats...
format HTML_BEGIN_CONTENTS =
<!------  @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<-------!>
$outfile
<!------  @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<-------!>
$title
<!------  Created by cinf.bat  -------!>
<HTML>
<HEAD>
<TITLE>@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< </TITLE>
$title
<table>
<tr>
<td><FONT COLOR="004040"><B>CINF : </B></FONT></td>
<td><FONT COLOR="004040">Copyright &copy;1996-97 Microsoft Corporation</FONT></td>
</tr>
<tr><td>	</td>
<td><FONT COLOR="004040">Version @<<<<<<<<<<< </FONT></td></tr>
$Version
</table>
<HR SIZE=3 COLOR="408080">
</HEAD>

<BODY>
.

format HTML_END =
<!--/FONT--!>
</BODY>
</HTML>
.

	open (INFILE, "$infile")	|| die "Can't open $infile for input\n";

	mkdir("htm", 0755);
	open (OUTFILE, ">$outfile") || die "Can't open $outfile for output\n";
	$oldhandle = select (OUTFILE);
	# restore outfile wrt directory htm
	$outfile =~ s/\\htm//;

	$title = "$INFClass INF: $infile";
	$Version = $CINFVersion'BuildVersion;
	$~ = "HTML_BEGIN_CONTENTS";
	write;

	print "<H2><A NAME=\"Top\">Contents of $infile</A></H2>\n";
	print "<H4><A HREF=\"#Errors\">Errors</A></H4>\n";
	print "<H4><A HREF=\"#Warnings\">Warnings</A></H4>\n";
	print "<H4><A HREF=\"#Index\">Index to sections</A></H4>\n";
	print "<H4><A HREF=\"#Source\">Source: $infile</A></H4>\n";
	print "<H4><A HREF=\"#Summary\">Summary of $infile</A></H4>\n";
	print "<H4><A HREF=\"summary.htm\">Back to summary page</A></H4>\n";
	print "<hr>\n";
	
	#print annotations.


	print "<A NAME=\"Errors\"><H2>Errors</H2></A>\n";
	($NumErrs) ? &print_err_index : print "No Errors.<P>\n" ;
	print "<A HREF=\"#Top\">To top of page</A>\n";
	print "<hr>\n";
	
	print "<A NAME=\"Warnings\"><H2>Warnings</H2></A>\n";
	($NumWrns) ? &print_wrn_index : print "No Warnings.<P>\n" ;
	print "<A HREF=\"#Top\">To top of page</A>\n";
	print "<hr>\n";
	
	print "<A NAME=\"Index\"><H2>Index to sections</H2></A>\n";
	&print_index;
	print "<A HREF=\"#Top\">To top of page</A>\n";
	print "<hr>\n";
	print "<A NAME=\"Source\"><H2>Source :$infile</H2></A>\n";
	&print_infile;
	print "<A HREF=\"#Top\">To top of page</A>\n";
	print "<hr>\n";
	print "<A NAME=\"Summary\"><H2>Summary of $infile</H2></A>\n";
	&summarize($infile, $outfile);

	if ( defined($SummaryProcedure) && $SummaryProcedure !~ /^$/)
		{ 
		&$SummaryProcedure; 	# calls INF specific summary procedure
	}

	print "<A HREF=\"#Top\">To top of page</A>\n";
	print "<hr>\n";

	$~ = "HTML_END";
	write;

	close (INFILE);
	close (OUTFILE);
	select ($oldhandle);

}


#-------------------------------------------------------------
#
# This sub prints the main page - summary.htm
#
# Globals changed:
#
# Usage:
#
#-------------------------------------------------------------

sub 
print_main_page
{
local($outfile, $file, $oldhandle, $title, $Version); 
format HTML_BEGIN =
<!------  @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<-------!>
$outfile
<!------  @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<-------!>
$title
<!------  Created by cinf.bat  -------!>
<HTML>
<HEAD>
<TITLE>@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< </TITLE>
$title
<table>
<tr>
<td><FONT COLOR="004040"><B>CINF : </B></FONT></td>
<td><FONT COLOR="004040">Copyright &copy;1996-97 Microsoft Corporation</FONT></td>
</tr>
<tr><td>	</td>
<td><FONT COLOR="004040">Version @<<<<<<<<<<< </FONT></td></tr>
$Version
</table>
<HR SIZE=3 COLOR="408080">
</HEAD>

<BODY>
.

format HTML_END =
<!--/FONT--!>
</BODY>
</HTML>
.

	$outfile = '.\htm\summary.htm';
	open (OUTFILE, ">$outfile") || die "Can't open $outfile for output\n";
	print "Summarizing in $outfile.\n";
	$oldhandle = select (OUTFILE);

	$title = "Summary of Checks on INF files";
	$Version = $CINFVersion'BuildVersion;
	$~ = "HTML_BEGIN";
	write;

	print "<H2>", $title, "</H2>\n";
	print "<pre>\n";
	print "Total files checked : $NumFiles\n";
	print "Files with errors   : $NumErrFiles\n";

	if ($PrintModemR0Statistics)
		{ print "Duplicate Rank0 ids : $NumR0Errs\n"; }

	print "</pre>\n";
	print "<hr>\n";

	# print details for each file.
	print "<br>\n";
	print "<table>\n";

	foreach $file (sort keys %Summary)
	{
		print "<tr>\n";
		print $Summary{$file}, "\n";
		print "</tr>\n";
	}
	print "</table>\n";
	print "<hr>\n";

	# print R0 check results.
	if ($PrintModemR0Statistics)
		{ ($NumR0Errs) ? &print_err_R0 : 
			print "<h3>No Rank0 conflicts found in files processed.</h3>\n" ;
	}
	print "<hr>\n";

	$~ = "HTML_END";
	write;

	close (OUTFILE);
	select ($oldhandle);
}


#############################################################
#
# subs used for printing.
#
#############################################################

#-------------------------------------------------------------
#
# This sub prints the index to sections in the Inf.
#
# Globals changed:
#
# Usage:
#
#-------------------------------------------------------------
sub 
print_index
{
local($sec, $key);

	foreach $sec (sort keys %SecStart) 
	{
		$key = $SecStart{$sec}; 
		print "<H5><A HREF=\"#Lnum$key\">[", $sec, "]</A></H5>\n";
	}
}

#-------------------------------------------------------------
#
# This sub prints the inf file with annotations - errs/wrns.
#
# Globals changed:
#
# Usage:
#
#-------------------------------------------------------------
sub 
print_infile
{

local($err, @errs);
	
	# print the infile here.
	print "<PRE>\n";

	while(<INFILE>)
	{
		#print each line with its ref as linenum.
		print "<A NAME=\"Lnum", $., "\">"; 
		chop if /\n$/;			 # removes \n
		s/</&lt/g; # so that htm interprets them right
		s/>/&gt/g;
		print;
		print "</A>\n";
		if (defined ($err = $StrErrs{$.})) 
		{
			print "<FONT COLOR=RED>";
			@errs = split ('; ', $err);
			foreach $err (@errs) 
			{
				next if ($err=~ /^$/);
				print ";";
				print "$err";
				print "\n";
			}
			print "</FONT>";
		}
		if (defined ($err = $StrWrns{$.})) 
		{
			print "<FONT COLOR=BLUE>";
			@errs = split ('; ', $err);
			foreach $err (@errs) 
			{
				next if ($err=~ /^$/);
				print ";";
				print "$err";
				print "\n";
			}
			print "</FONT>";
		}
	}

	print "</PRE>\n";
}

			
#------------------------------------------------------------
# This sub prints the error index.
#
# Usage: 
#		print_err_index
#------------------------------------------------------------

sub print_err_index
{
local(@err_lines, @errs, $key, $err);

	print "Total errors: $NumErrs\n";
	print "<UL>\n";
	# get all line nums with errors. sort them and print out the errors.
	@err_lines=();
	foreach $key (keys(%StrErrs)) 
	{
		@err_lines = (@err_lines, $key);
	}
	@err_lines = sort numeric @err_lines;
	foreach $key (@err_lines) 
	{
		$err = $StrErrs{$key}; 
		@errs = split ('; ', $err);
		foreach $err (@errs) 
		{
			next if ($err=~ /^$/);
			print "<LI> <A HREF=\"#Lnum$key\">Line$key: $err</A>\n";
		}
	}
	print "</UL>\n";
}

#-------------------------------------------------------------
#
# This sub prints the warning index.
#
# Globals changed:
#
# Usage:
#
#-------------------------------------------------------------
sub print_wrn_index
{
local(@err_lines, @errs, $key, $err);

	print "Total warnings: $NumWrns\n";
	print "<UL>\n";
	# get all line nums with errors. sort them and print out the errors.
	@err_lines=();
	foreach $key (keys(%StrWrns)) 
	{
		@err_lines = (@err_lines, $key);
	}
	@err_lines = sort numeric @err_lines;
	foreach $key (@err_lines) 
	{
		$err = $StrWrns{$key}; 
		@errs = split ('; ', $err);
		foreach $err (@errs) 
		{
			next if ($err=~ /^$/);
			print "<LI> <A HREF=\"#Lnum$key\">Line$key: $err</A>\n";
		}
	}
	print "</UL>\n";
}


#-------------------------------------------------------------
#
# Checks for the correct version
#
#-------------------------------------------------------------
sub
VersionChecking
{
   local ($Diff);

   $Diff = 0;
   if ( defined($ADDREGVersion'BuildVersion) &&
	$CINFVersion'BuildVersion ne $ADDREGVersion'BuildVersion) 
	   { print "\"addreg.pl\" build version $ADDREGVersion'BuildVersion\n";
		 $Diff = 1; }
   if ( defined($PRSECVersion'BuildVersion) &&
	$CINFVersion'BuildVersion ne $PRSECVersion'BuildVersion)
	   { print "\"prsec.pl\" build version $PRSECVersion'BuildVersion\n";
		 $Diff = 1; }
   if ( defined($INFPARSEVersion'BuildVersion) &&
	$CINFVersion'BuildVersion ne $INFPARSEVersion'BuildVersion) 
	   { print "\"InfParse.pl\" build version $INFPARSEVersion'BuildVersion\n";
		 $Diff = 1; }

	if ( $Diff != 0)
	{
	print "Build versions differs. Continue (Y/N) [Y]?";
	$VersionDiff = <STDIN>;
	chop $VersionDiff;
	die "\n" if ( (uc $VersionDiff) =~ /^N/ );
	}
}

__END__

echo off
:endofperl
@perl %0.bat %1 %2 %3 %4 %5 %6 %7 %8 %9
